from collections import deque

from ..query import QueryVisitor, Link, Field
from ..types import TypeRefMeta, OptionalMeta, SequenceMeta, get_type


class Denormalize(QueryVisitor):

    def __init__(self, graph, result):
        self._types = graph.__types__
        self._result = result
        self._type = deque([self._types['__root__']])
        self._data = deque([result])
        self._res = deque()

    def process(self, query):
        assert not self._res, self._res
        self._res.append({})
        self.visit(query)
        return self._res.pop()

    def visit_field(self, obj: Field):
        self._res[-1][obj.result_key] = self._data[-1][obj.result_key]

    def visit_link(self, obj: Link):
        type_ = self._type[-1].__field_types__[obj.name]
        if isinstance(type_, TypeRefMeta):
            self._type.append(get_type(self._types, type_))
            self._res.append({})
            self._data.append(self._data[-1][obj.result_key])
            super().visit_link(obj)
            self._data.pop()
            self._res[-1][obj.result_key] = self._res.pop()
            self._type.pop()
        elif isinstance(type_, SequenceMeta):
            self._type.append(get_type(self._types, type_.__item_type__))
            items = []
            for item in self._data[-1][obj.result_key]:
                self._res.append({})
                self._data.append(item)
                super().visit_link(obj)
                self._data.pop()
                items.append(self._res.pop())
            self._res[-1][obj.result_key] = items
            self._type.pop()
        elif isinstance(type_, OptionalMeta):
            if self._data[-1][obj.result_key] is None:
                self._res[-1][obj.result_key] = None
            else:
                self._type.append(get_type(self._types, type_.__type__))
                self._res.append({})
                self._data.append(self._data[-1][obj.result_key])
                super().visit_link(obj)
                self._data.pop()
                self._res[-1][obj.result_key] = self._res.pop()
                self._type.pop()
        else:
            raise AssertionError(repr(type_))



import inspect

from asyncio import wait, FIRST_COMPLETED, gather, CancelledError
from asyncio import get_event_loop


class AsyncIOExecutor:

    def __init__(self, loop=None):
        self._loop = loop or get_event_loop()

    def submit(self, fn, *args, **kwargs):
        coro = fn(*args, **kwargs)
        if not inspect.isawaitable(coro):
            raise TypeError('{!r} returned non-awaitable object {!r}'
                            .format(fn, coro))
        return self._loop.create_task(coro)

    async def process(self, queue, workflow):
        try:
            while queue.__futures__:
                done, _ = await wait(queue.__futures__,
                                     return_when=FIRST_COMPLETED)
                queue.progress(done)
            return workflow.result()
        except CancelledError:
            for task in queue.__futures__:
                task.cancel()
            await gather(*queue.__futures__)
            raise
